package atomicstryker.minions.common.jobmanager;

import java.util.ArrayList;
import java.util.List;

import net.minecraft.block.Block;
import net.minecraft.block.material.Material;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.init.Blocks;
import net.minecraft.item.ItemStack;
import net.minecraft.util.BlockPos;
import net.minecraft.util.MathHelper;
import net.minecraft.world.World;
import atomicstryker.astarpathing.AStarNode;
import atomicstryker.astarpathing.AStarStatic;
import atomicstryker.minions.common.MinionsCore;
import atomicstryker.minions.common.entity.EntityMinion;

/**
 * Blocktask super schematic. By default a Blocktask doesnt change the Block.
 * 
 * 
 * @author AtomicStryker
 */

public abstract class BlockTask
{
    protected final Minion_Job_Manager boss;
    public final int posX;
    public final int posY;
    public final int posZ;
    public final BlockPos pos;
    private boolean startedTask;
    protected EntityMinion worker;
    protected double accessRangeSq;
    protected long taskDurationMillis;
    public boolean workerReachedBlock;
    protected long timeBlockReached;
    private long taskTimeStarted;
    private double startMinionX;
    private double startMinionZ;

    /**
     * Creates a new BlockTast instance
     * 
     * @param boss
     *            Task Manager to notify about events coming from here
     * @param input
     *            worker to be assigned to the task, CAN be null to just store
     *            the job for later. Task will not start until a worker is
     *            assigned.
     * @param ix
     *            Block x coordinate
     * @param iy
     *            Block y coordinate
     * @param iz
     *            Block z coordinate
     */
    public BlockTask(Minion_Job_Manager boss, EntityMinion input, int ix, int iy, int iz)
    {
        // System.out.println("BlockTask created!");
        this.boss = boss;
        this.worker = input;
        this.posX = ix;
        this.posY = iy;
        this.posZ = iz;
        pos = new BlockPos(ix, iy, iz);
        startedTask = false;
        accessRangeSq = 9.0D;
        taskDurationMillis = 1000L;
        workerReachedBlock = false;
    }

    /**
     * Assigns a worker to this Task. The worker should forward Update Ticks to
     * this task.
     * 
     * @param input
     *            worker to be assigned
     */
    public void setWorker(EntityMinion input)
    {
        MinionsCore.debugPrint("task "+this+" at ["+posX+"|"+posY+"|"+posZ+"] assigned to worker "+worker);
        this.worker = input;
    }

    /**
     * Specifies from which "block reach" range this task can be done
     * 
     * @param input
     */
    public void setAccessRange(double input)
    {
        this.accessRangeSq = input;
    }

    /**
     * Specifies the task's work duration in millis
     * 
     * @param input
     */
    public void setTaskDuration(long input)
    {
        this.taskDurationMillis = input;
    }

    /**
     * Update tick coming from the assigned worker. Hence, does not get called
     * when there is no worker assigned.
     */
    public void onUpdate()
    {
        if (!startedTask)
        {
            if (!worker.inventoryFull)
            {
                onStartedTask();
            }
            else
            {
                worker.returningGoods = true;
                worker.runInventoryDumpLogic();
                MinionsCore.debugPrint("Blocktask "+this+" worker "+worker+" is full, sending to return goods");
            }
        }
        else if (!workerReachedBlock && System.currentTimeMillis() - taskTimeStarted > 1000L)
        {
            if (Math.abs(startMinionX - worker.field_70165_t) < 1D && Math.abs(startMinionZ - worker.field_70161_v) < 1D)
            {
                onWorkerPathFailed();
            }
            else
            {
                taskTimeStarted = System.currentTimeMillis();
                startMinionX = worker.field_70165_t;
                startMinionZ = worker.field_70161_v;
            }
        }

        if (isWorking())
        {
            worker.faceBlock(posX, posY, posZ);
            worker.func_70096_w().func_75692_b(16, Integer.valueOf(1));
            worker.func_70096_w().func_75692_b(17, Integer.valueOf(posX));
            worker.func_70096_w().func_75692_b(18, Integer.valueOf(posY));
            worker.func_70096_w().func_75692_b(19, Integer.valueOf(posZ));
        }

        if (!workerReachedBlock)
        {
            if (isEntityInAccessRange(worker))
            {
                onReachedTaskBlock();
            }
        }
        else if ((System.currentTimeMillis() - timeBlockReached) > (taskDurationMillis / worker.workSpeed))
        {
            onFinishedTask();
        }
    }

    /**
     * Causes the assigned worker to attempt to navigate the next pre-computed
     * path. If all paths failed, as a last resort the worker is teleported to
     * the tasked Block.
     */
    public void onWorkerPathFailed()
    {
        // System.out.println("BlockTask onWorkerPathFailed all paths failed, teleporting dat minion");
        worker.performTeleportToTarget();
        onReachedTaskBlock();
    }

    /**
     * Called when a worker comes within accessing range of the Block. Commences
     * animation, sets starting time, stops movement.
     */
    public void onReachedTaskBlock()
    {
        // System.out.println("BlockTask onReachedTaskBlock");
        workerReachedBlock = true;
        timeBlockReached = System.currentTimeMillis();
        this.worker.setWorking(true);
        //this.worker.setPathToEntity(null);

        worker.adaptItem(worker.field_70170_p.func_180495_p(new BlockPos(posX, posY, posZ)).func_177230_c().func_149688_o());
    }

    /**
     * Called when the BlockTask got a worker assigned and should commence
     * execution. Starts path computing, aborts if no path can be found.
     */
    public void onStartedTask()
    {
        if (startedTask)
            return;
        startedTask = true;
        MinionsCore.debugPrint("onStartedTask "+this+" ["+this.posX+"|"+this.posY+"|"+this.posZ+"], worker "+worker);

        taskTimeStarted = System.currentTimeMillis();
        startMinionX = worker.field_70165_t;
        startMinionZ = worker.field_70161_v;
        
        AStarNode[] possibleAccessNodes = getAccessNodesSorted(MathHelper.func_76128_c(worker.field_70165_t), MathHelper.func_76128_c(worker.field_70163_u) - 1, MathHelper.func_76128_c(worker.field_70161_v));
        if (possibleAccessNodes.length != 0)
        {
            this.worker.orderMinionToMoveTo(possibleAccessNodes, false);
        }
        else
        {
            MinionsCore.debugPrint("Teleporting Minion to impathable task "+this);
            worker.performTeleportToTarget();
        }
    }

    /**
     * Called when the deed is done. Resets the worker. Notifies the
     * taskmanager.
     */
    public void onFinishedTask()
    {
        MinionsCore.debugPrint("onFinishedTask "+this+" ["+this.posX+"|"+this.posY+"|"+this.posZ+"], resetting minion "+worker);
        this.worker.giveTask(null, true);
        
        if (boss != null)
        {
            boss.onTaskFinished(this, posX, posY, posZ);
        }
    }

    /**
     * @return true when the worker has moved within reach distance of the
     *         target Block, false otherwise
     */
    public boolean isWorking()
    {
        return workerReachedBlock;
    }

    /**
     * @param ent
     * @return true when the entity currently is within reach distance of the
     *         target Block, false otherwise
     */
    private boolean isEntityInAccessRange(EntityLivingBase ent)
    {
        return (ent.func_70092_e(this.posX, this.posY, this.posZ) < accessRangeSq);
    }

    /**
     * @param workerX
     * @param workerY
     * @param workerZ
     * @return an Array of pathable AStarNodes, starting with the closest one to
     *         parameter coordinates and ascending. Array can be size 0 but is
     *         != null
     */
    private AStarNode[] getAccessNodesSorted(int workerX, int workerY, int workerZ)
    {
        return AStarStatic.getAccessNodesSorted(worker.field_70170_p, posX, posY, posZ);
    }

    /**
     * Figures out what ItemStack would result from breaking a Block in the
     * World
     */
    protected List<ItemStack> getItemStacksFromWorldBlock(World world, int i, int j, int k)
    {
        Block block = world.func_180495_p(new BlockPos(i, j, k)).func_177230_c();
        Material m = block.func_149688_o();
        
        if (block == Blocks.field_150350_a || m == Material.field_151586_h || m == Material.field_151587_i || m == Material.field_151584_j || m == Material.field_151585_k)
        {
            return new ArrayList<ItemStack>();
        }
        
        BlockPos pos = new BlockPos(i, j, k);
        IBlockState state = world.func_180495_p(pos);
        return block.getDrops(world, pos, state, state.func_177230_c().func_176201_c(state));
    }

    protected void putBlockHarvestInWorkerInventory(List<ItemStack> stackList)
    {
        if (stackList != null)
        {
            for (int i = 0; i < stackList.size(); i++)
            {
                if (!this.worker.inventory.addItemStackToInventory(stackList.get(i)))
                {
                    worker.inventoryFull = true;
                    worker.field_70170_p.func_72838_d(new EntityItem(worker.field_70170_p, this.posX, this.posY, this.posZ, stackList.get(i)));
                }
            }
        }
    }

    @Override
    public boolean equals(Object o)
    {
        if (o instanceof BlockTask)
        {
            BlockTask checktask = (BlockTask) o;
            return (this.posX == checktask.posX && this.posY == checktask.posY && this.posZ == checktask.posZ);
        }
        return false;
    }
}
